/* QVM86 virtualisation module for qemu

   Copyright (C) 2005 CodeSourcery, LLC.
   Written by Paul Brook.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   Assembly code for switching between host and monitor.
   This also implements all the monitor functionality.
   This code will be relocated at runtime, so must be PIC.
 */
/* The context switch works as follows:
TODO: update this comment.

   Save general regs on the stack.
   Save CR and descriptor table pointer in the shared area.
   Switch pagetables.  All the data we need should be mapped at the same
   addresses in both VMs.
   Load the new descriptor table pointers.
   Switch to the new stack, code and data segents.
   Load general regs off the stack.  */

#ifdef _WIN32
#define NAME(x) _##x
#else
#define NAME(x) x
#endif

	.text
	.global NAME(qvm86_monitor_start)
NAME(qvm86_monitor_start):

#define FIELD_S(s, o, n) .equ O_##s##_##n, o;
#define FIELD_L(s, o, n) FIELD_S(s, o, n)
#define FIELD_END(o) .zero o
#include "monitor-env.h"

/* Switch from host to monitor.  */
	.global NAME(qvm86_monitor_run)
#ifndef _WIN32
	.type NAME(qvm86_monitor_run), @function
#endif
NAME(qvm86_monitor_run):
	/* Load the address of the shared data block off the stack.  */
	movl 4(%esp), %eax
	/* Save state on the host stack.  */
	pushfl
	pushal
	pushl %ds
	pushl %es
	pushl %fs
	pushl %gs

	movl %eax, %ebx

	/* Clear the interrupt flag.  We saved the eflags register earlier,
	   so this will be restored eventually.  */
	cli

	/* Set cs value of the return vector.  */
	movw %cs, %ax
	movw %ax, O_host_cs(%ebx)
	
	/* Record host state in the shared monitor page.  */
	sgdt O_host_gdt_limit(%ebx)
	sidt O_host_idt_limit(%ebx)
	sldt O_host_ldt(%ebx)
	str O_host_tr(%ebx)
	
	movw %ds, O_host_ds(%ebx)
	movw %ss, O_host_ss(%ebx)
	movl %esp, O_host_esp(%ebx)

	/* Map this page into monitor space.  */
	movl O_host_switch_ptr(%ebx), %eax
	movl O_host_switch_pde_low(%ebx), %edi
	xchgl %edi, (%eax)
        movl %edi, O_host_switch_pde_low(%ebx)
        movl O_host_pae(%ebx), %ecx
        testl %ecx, %ecx
        jz 1f
	movl O_host_switch_pde_high(%ebx), %edi
	xchgl %edi, 4(%eax)
        movl %edi, O_host_switch_pde_high(%ebx)
1:
	
	movl %cr0, %eax
	movl %cr2, %ecx
	movl %cr3, %edx
	movl %cr4, %esi
	movl %eax, O_host_cr0(%ebx)
	movl %ecx, O_host_cr2(%ebx)
	movl %edx, O_host_cr3(%ebx)
	movl %esi, O_host_cr4(%ebx)

	/* Load the monitor CRx values.  This switches to the monitor address
	   space.  */
	movl O_mon_cr0(%ebx), %eax
	movl O_mon_cr4(%ebx), %ecx
	movl O_mon_cr3(%ebx), %edx
	movl %edx, %cr3
	movl %eax, %cr0
	movl %ecx, %cr4
	movl %edx, %cr3
	jmp 1f
1:
	/* Load monitor gdt, itd and ldt.  */
	lgdt O_mon_gdt_limit(%ebx)
	lidt O_mon_idt_limit(%ebx)
	lldt O_mon_ldt(%ebx)

	/* Switch to the monitor stack and code segment.  */
	lss O_mon_esp(%ebx), %esp
	ljmp *O_mon_eip(%ebx)

	/* We end up here when entering the monitor from the host.  */
	.global NAME(qvm86_monitor_entry)
NAME(qvm86_monitor_entry):
	/* Load the data segment selector.  */
	movw %ss, %ax
	movw %ax, %ds
	movw %ax, %es

	/* The data will be mapped at address zero.  */
	xorl %ebx, %ebx

	/* Load the task register.  */
	ltr O_mon_tr(%ebx)

	/* Restore the original page mapping.  */
	movl O_mon_switch_ptr(%ebx), %eax
	movl O_host_switch_pde_low(%ebx), %edi
	xchgl %edi, (%eax)
        movl %edi, O_host_switch_pde_low(%ebx)
        movl O_host_pae(%ebx), %ecx
        testl %ecx, %ecx
        jz 1f
	movl O_host_switch_pde_high(%ebx), %edi
	xchgl %edi, 4(%eax)
        movl %edi, O_host_switch_pde_high(%ebx)
1:
	/* Invalidate the TLB entry.  */
	movw O_mon_linmap(%ebx), %ax
	movw %ax, %fs
	movl O_host_env(%ebx), %eax
	invlpg %fs:(%eax)

return_to_guest:
	/* Restore regs, and return to the monitor.  */
	pop %gs
	pop %fs
	pop %es
	pop %ds
	popal
	popfl
#if 0
	pushl %ss
	pushl %esp
	pushfl
	pushl %cs
	call 1f
1:
	pushl $0
	pushl $0
	jmp qvm86_monitor_faulthandler
#endif

	ret

/* Switch from monitor to host.  */
	.global NAME(qvm86_return_to_host)
#ifndef _WIN32
	.type NAME(qvm86_return_to_host), @function
#endif
NAME(qvm86_return_to_host):
	/* Save register state on the stack.  */
	pushfl
	pushal
	push %ds
	push %es
	push %fs
	push %gs

	/* Switch to the monitor data segment.  */
	movw %ss, %ax
	movw %ax, %ds
	movw %ax, %es

	/* The monitor_env structure will be mapped at data address zero.  */
	xorl %ebx, %ebx

	/* Save monitor CRx.  */
	movl %cr0, %eax
	movl %cr4, %ecx
	movl %cr3, %edx
	movl %cr2, %esi
	movl %eax, O_mon_cr0(%ebx)
	movl %ecx, O_mon_cr4(%ebx)
	movl %edx, O_mon_cr3(%ebx)
	movl %esi, O_mon_cr2(%ebx)

	/* Save monitor gdt, itd and ldt.  */
	sgdt O_mon_gdt_limit(%ebx)
	sidt O_mon_idt_limit(%ebx)
	sldt O_mon_ldt(%ebx)

	/* Save monitor stack pointer.  */
	movl %esp, O_mon_esp(%ebx)
	movw %ss, %ax
	movw %ax, O_mon_ss(%ebx)

	/* Load linear mapped segment.  */
	movw O_mon_linmap(%ebx), %ax
	movw %ax, %fs

	/* Record the reason for returning.  This is the vector saved
	   on the stack by the handler stub.  At this point the stack
	   holds 4 segment descriptors, 8 general regs, return address,
	   flag word, vector, error code and interrupt frame.  */
	movl 56(%esp), %ecx
	movl %ecx, O_mon_reason(%ebx)
	movl 60(%esp), %eax
	movl %eax, O_mon_errcode(%ebx)

/* This code can't handle PAE.  */
#if 0
	cmp $14, %ecx
	jne 1f
	/* We can handle some page faults without needing to switch back to
	   the host.  */
	call do_pagefault
	cmp $0, %eax
	jne return_to_guest
1:
#endif
#if 0
/* This is disabled because it can cause host reboots when CS:EIP is not
   readable.  */
	cmp $13, %ecx
	jne 1f
	/* Point eax at the code location.  */
	leal 64(%esp), %eax
	call do_gpf
1:
#endif

	/* Map in the shared page.  */
	movl O_mon_switch_ptr(%ebx), %eax
	movl O_host_switch_pde_low(%ebx), %edi
	xchgl %edi, (%eax)
        movl %edi, O_host_switch_pde_low(%ebx)
        movl O_host_pae(%ebx), %ecx
        testl %ecx, %ecx
        jz 1f
	movl O_host_switch_pde_high(%ebx), %edi
	xchgl %edi, 4(%eax)
        movl %edi, O_host_switch_pde_high(%ebx)
1:
	/* Load the shared data pointer, and invalidate the TLB entry.  */
	movl O_host_env(%ebx), %ebx
	invlpg %fs:(%ebx)

	/* Switch to using shared data page.  */
	movw %fs, %ax
	movw %ax, %ds

	/* Jump to code in the shared page so it doesn't disappear when we
	   switch address spaces.  */
	ljmp *O_mon_switch_eip(%ebx)

	/* We jump here after mapping the shared page, but still in monitor
	   space.  */
	.global NAME(qvm86_monitor_switch)
NAME(qvm86_monitor_switch):
	/* Load the host CRx values.  */
	movl O_host_cr0(%ebx), %eax
	movl O_host_cr2(%ebx), %ecx
	movl O_host_cr4(%ebx), %edx
	movl O_host_cr3(%ebx), %esi
	movl %eax, %cr0
	movl %ecx, %cr2
	movl %edx, %cr4
	movl %esi, %cr3
	jmp 1f
1:
	/* Load the host gdt, idt and ldt.  */
	lgdt O_host_gdt_limit(%ebx)
	lidt O_host_idt_limit(%ebx)
	lldt O_host_ldt(%ebx)

	/* Switch to host stack and code segment.  */
	lss O_host_esp(%ebx), %esp
	ljmp *O_host_eip(%ebx)

	/* We end up here when returning from the monitor to the host.  */
	.global NAME(qvm86_monitor_return)
NAME(qvm86_monitor_return):
	/* Restore host segment descriptors.  */
	popl %gs
	popl %fs
	popl %es
	popl %ds

	/* Clear the busy bit in the descriptor.  */
	movzwl O_host_tr(%ebx), %eax
	andl $0xfffffff8, %eax
	addl O_host_gdt_base(%ebx), %eax
	/* TODO: cope with nonzero data segment offsets.  */
	andl $0xfffffdff, 4(%eax)
	/* Switch back to the host task.  */
	ltr O_host_tr(%ebx)

	/* Restore the original guest page mapping.  */
	movl O_host_switch_ptr(%ebx), %eax
	movl O_host_switch_pde_low(%ebx), %edi
	xchgl %edi, (%eax)
        movl %edi, O_host_switch_pde_low(%ebx)
        movl O_host_pae(%ebx), %ecx
        testl %ecx, %ecx
        jz 1f
	movl O_host_switch_pde_high(%ebx), %edi
	xchgl %edi, 4(%eax)
        movl %edi, O_host_switch_pde_high(%ebx)
1:
	
	/* Restore host general regs, and return.  */
	popal
	popfl
	ret

	.global NAME(qvm86_start_guest)
NAME(qvm86_start_guest):
	iret

/* Fault handler vector.  Stack already contains eflags, ss, eip,
   error code and vector.  */
	.global NAME(qvm86_monitor_faulthandler)
NAME(qvm86_monitor_faulthandler):
	call NAME(qvm86_return_to_host)
	/* pop the saved error and vector.  */
	addl $8, %esp
	iret

/* Interrupt handler vector.  Stack already contains eflags, ss, eip,
   error code and vector.  */
	.global NAME(qvm86_monitor_inthandler)
NAME(qvm86_monitor_inthandler):
	call NAME(qvm86_return_to_host)
	/* pop the saved error and vector.  */
	addl $8, %esp
	iret

#if 0
do_pagefault:
	/* Try to handle a page fault.  return %eax=0 if we succeeded.
	    This is a somewhat simplified version of qvm86_shadow_fault.  */
	/* %eax contains the error code.  */
	testl $0x9, %eax
	jz 1f
	/* Fault is a protection violation.  */
	/* ??? Handle write faults on pages that wetre mapped readonly.  */
	xorl %eax, %eax
	ret
1:
	/* Fault is an absent page.  */

	/* Get the fault address.  */
	movl O_mon_cr2(%ebx), %ebp
	andl $0xfffff000, %ebp
	/* Set the RW bit on the fault address.  */
	andl $2, %eax
	orl %eax, %ebp

	/* Find the guest PDE.  */
	movl %ebp, %ecx
	shrl $22, %ecx
	movl O_mon_guest_pgd(%ebx), %esi
	movl (%esi, %ecx, 4), %edx
	test $1, %edx
	jz 2f
	/* Is this a user page.  */
	testl $4, %edx
	jz 2f
	/* Check for writing to a readonly page.  */
	testl $2, %edx
	jnz 1f
	testl $2, O_mon_errcode(%ebx)
	jz 1f
2:
	/* A real guest fault.  Bounce to host.  */
	xorl %eax, %eax
	ret
1:
	/* The guest PDE access bit will already have been set when the
	   shadow PDE was created.  */
	movl %edx, %eax
	andl $0xfffff000, %eax
	/* Map in the guest pagetable.  */
	call map_guest_pagetable
	cmpl $0, %eax
	jne 1f
	/* guest pagetable not mapped in.  Bounce to host.  */
	/* %eax is already zero.  */
	ret
1:
	movl O_mon_shadow_pgd(%ebx), %esi
	movl (%esi, %ecx, 4), %edi
	test $1, %edi
	jnz 1f
	/* TODO: Is it worth the hassle to handle this here?  There are
	   probably other host shadow structures that need keeping up to
	   date.  */
	/* Shadow page directory does not exist.  Bounce to host.  */
	xorl %eax, %eax
	ret
1:
	/* Now deal with the pagetable entry.  */
	/* Get the guest pagetable entry.  */
	movl %ebp, %eax
	shrl $12, %eax
	andl $0x3ff, %eax
	movl O_mon_free_page(%ebx), %esi
	movl (%esi, %eax, 4), %edx

	/* Is guest PTE present.  */
	testl $1, %edx
	jz 2f
	/* Is this a user page.  */
	testl $4, %edx
	jz 2f
	/* Check for writing to a readonly page.  */
	testl $2, %edx
	jnz 1f
	testl $2, O_mon_errcode(%ebx)
	jz 1f
2:
	/* A real guest fault.  Bounce to host.  */
	xorl %eax, %eax
	ret
1:
	/* See if this is a read or write fault.  */
	movl O_mon_errcode(%ebx), %edi
	andl $2, %edi
	jz 1f
	/* This is a write fault.  Set the dirty bit.  */
	orl $0x40, %edx
1:
	/* Set the accessed bit, and write the guest pagetable back.  */
	movl (%esi, %eax, 4), %edx
	movl %edx, (%esi, %eax, 4)

	/* Clear the pagetable write premission bit for read faults.  */
	notl %edi
	andl %edi, %edx

	movl %edx, %eax
	shrl $12, %eax
	/* Get the host physical address of this page.  */
	movl O_mon_physmem_map(%ebx), %esi
	movl %fs:(%esi, %eax, 4), %ecx
	testl %ecx, %ecx
	jnz 1f
	/* Guest page not mapped in.  Bounce to host.  */
	xorl %eax, %eax
	ret
1:
	/* We have everything we need.  Set the shadow pagetable entry.  */
	movl %ebp, %eax
	movl %eax, O_mon_foo(%ebx)
	shrl $12, %eax
	movl O_mon_shadow_pte(%ebx), %esi
	andl $0x19f, %edx
	orl %ecx, %edx
	movl %edx, O_mon_foo2(%ebx)
	movl %fs:(%esi, %eax, 4), %ecx
	testl $1, %ecx
	jz 1f
	/* Page already shadowed.  */
	xorl %eax, %eax
	ret
1:
	movl %edx, %fs:(%esi, %eax, 4)
	/* Invalidate the TLB entry for this page.  */
	movl %ebp, %eax
	invlpg (%eax)

	/* Return to the caller, letting them know we handled the fault.  */
	incl O_mon_fault_count(%ebx)
	movl $1, %eax
	ret

	/* Map a guest physical address (%eax) into monitor space.
	  Sets %eax to the physical address of the page, or zero if the page
	  is not available.  */
map_guest_pagetable:
	pushl %esi
	pushl %ecx
	/* Get the host physical address.  */
	shrl $12, %eax
	movl O_mon_physmem_map(%ebx), %esi
	movl %fs:(%esi, %eax, 4), %eax
	cmp $0, %eax
	je 1f
	/* Map the page into monitor space.  */
	orl $3, %eax
	movl O_mon_free_pte(%ebx), %ecx
	movl %eax, (%ecx)
	movl O_mon_free_page(%ebx), %ecx
	invlpg (%ecx)
1:
	popl %ecx
	popl %esi
	andl $0xfffff000, %eax
	ret

#endif
/* Special handling for General Protection Faults.  We're mainly interested
   in INT n instructions so that we can pass these back to userspace.
   eax points to the instruction that caused the fault.  */
do_gpf:
	/* TODO: This could cause an exception if the current instruction
	   is outside the CS segment limit, or overlaps a page boundary.  */
	/* Load fs with the guest code segment.  */
	movw 4(%eax), %fs
	movl (%eax), %esi
	movzbl %fs:(%esi), %edx
	cmpl $0xcd, %edx
	jne 1f
	/* The faulting instruction was an INT n instruction.  */
	movzbl %fs:1(%esi), %edx
	/* Tell the host to do the interrupt.  */
	orl $0x200, %edx
	movl %edx, O_mon_reason(%ebx)
	movl %esi, (%eax)
1:
	ret
	.global NAME(qvm86_monitor_end)
NAME(qvm86_monitor_end):
